/* Gnome-Streamer
 * Copyright (C) <2001> Richard Boulton <richard-gst@tartarus.org>
 *
 * Based on example.c:
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "artsdsink.h"

/* elementfactory information */
static GstElementDetails artsdsink_details = {
  "aRtsd audio sink",
  "Sink/Artsdsink",
  "Plays audio to an aRts server",
  VERSION,
  "Richard Boulton <richard-gst@tartarus.org>",
  "(C) 2001",
};

/* Signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_MUTE,
  ARG_DEPTH,
  ARG_CHANNELS,
  ARG_RATE,
  ARG_NAME,
};

static GstPadTemplate*
sink_factory (void)
{
  return 
    gst_padtemplate_new (
  	"sink",					/* the name of the pads */
  	GST_PAD_SINK,				/* type of the pad */
  	GST_PAD_ALWAYS,				/* ALWAYS/SOMETIMES */
  	gst_caps_new (
     	  "artsdsink_sink",				/* the name of the caps */
     	  "audio/raw",				/* the mime type of the caps */
	  gst_props_new (
     	    "format",     GST_PROPS_STRING ("int"),
     	    "law",        GST_PROPS_INT (0),
     	    "endianness", GST_PROPS_INT (G_BYTE_ORDER),
     	    "signed",     GST_PROPS_BOOLEAN (FALSE),
     	    "width",      GST_PROPS_INT (8),
     	    "depth",      GST_PROPS_INT (8),
     	    "rate",       GST_PROPS_INT_RANGE (8000, 96000),
     	    "channels",   GST_PROPS_LIST (GST_PROPS_INT (1), GST_PROPS_INT (2)),
	    NULL)
  	),
  	gst_caps_new (
     	  "artsdsink_sink",				/* the name of the caps */
     	  "audio/raw",				/* the mime type of the caps */
	  gst_props_new (
     	    "format",     GST_PROPS_STRING ("int"),
     	    "law",        GST_PROPS_INT (0),
     	    "endianness", GST_PROPS_INT (G_BYTE_ORDER),
     	    "signed",     GST_PROPS_BOOLEAN (TRUE),
     	    "width",      GST_PROPS_INT (16),
     	    "depth",      GST_PROPS_INT (16),
     	    "rate",       GST_PROPS_INT_RANGE (8000, 96000),
     	    "channels",   GST_PROPS_LIST (GST_PROPS_INT (1), GST_PROPS_INT (2)),
	    NULL)
  	),
  	NULL);
}

static void			gst_artsdsink_class_init		(GstArtsdsinkClass *klass);
static void			gst_artsdsink_init		(GstArtsdsink *artsdsink);

static gboolean			gst_artsdsink_open_audio		(GstArtsdsink *sink);
static void			gst_artsdsink_close_audio		(GstArtsdsink *sink);
static GstElementStateReturn	gst_artsdsink_change_state	(GstElement *element);
static gboolean			gst_artsdsink_sync_parms		(GstArtsdsink *artsdsink);

static void			gst_artsdsink_chain		(GstPad *pad, GstBuffer *buf);

static void			gst_artsdsink_set_arg		(GtkObject *object, GtkArg *arg, guint id);
static void			gst_artsdsink_get_arg		(GtkObject *object, GtkArg *arg, guint id);

static GstPadTemplate *sink_template;

#define GST_TYPE_ARTSDSINK_DEPTHS (gst_artsdsink_depths_get_type())
static GtkType
gst_artsdsink_depths_get_type (void)
{
  static GtkType artsdsink_depths_type = 0;
  static GtkEnumValue artsdsink_depths[] = {
    {8, "8", "8 Bits"},
    {16, "16", "16 Bits"},
    {0, NULL, NULL},
  };
  if (!artsdsink_depths_type) {
    artsdsink_depths_type = gtk_type_register_enum("GstArtsdsinkDepths", artsdsink_depths);
  }
  return artsdsink_depths_type;
}

#define GST_TYPE_ARTSDSINK_CHANNELS (gst_artsdsink_channels_get_type())
static GtkType
gst_artsdsink_channels_get_type (void)
{
  static GtkType artsdsink_channels_type = 0;
  static GtkEnumValue artsdsink_channels[] = {
    {1, "1", "Mono"},
    {2, "2", "Stereo"},
    {0, NULL, NULL},
  };
  if (!artsdsink_channels_type) {
    artsdsink_channels_type = gtk_type_register_enum("GstArtsdsinkChannels", artsdsink_channels);
  }
  return artsdsink_channels_type;
}


static GstElementClass *parent_class = NULL;
//static guint gst_artsdsink_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_artsdsink_get_type (void)
{
  static GtkType artsdsink_type = 0;

  if (!artsdsink_type) {
    static const GtkTypeInfo artsdsink_info = {
      "GstArtsdsink",
      sizeof(GstArtsdsink),
      sizeof(GstArtsdsinkClass),
      (GtkClassInitFunc)gst_artsdsink_class_init,
      (GtkObjectInitFunc)gst_artsdsink_init,
      (GtkArgSetFunc)gst_artsdsink_set_arg,
      (GtkArgGetFunc)gst_artsdsink_get_arg,
      (GtkClassInitFunc)NULL,
    };
    artsdsink_type = gtk_type_unique(GST_TYPE_ELEMENT,&artsdsink_info);
  }
  return artsdsink_type;
}

static void
gst_artsdsink_class_init (GstArtsdsinkClass *klass)
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtk_object_add_arg_type("GstArtsdsink::mute", GTK_TYPE_BOOL,
			  GTK_ARG_READWRITE, ARG_MUTE);
  gtk_object_add_arg_type("GstArtsdsink::depth", GST_TYPE_ARTSDSINK_DEPTHS,
			  GTK_ARG_READWRITE, ARG_DEPTH);
  gtk_object_add_arg_type("GstArtsdsink::channels", GST_TYPE_ARTSDSINK_CHANNELS,
			  GTK_ARG_READWRITE, ARG_CHANNELS);
  gtk_object_add_arg_type("GstArtsdsink::frequency", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_RATE);
  gtk_object_add_arg_type("GstArtsdsink::name", GTK_TYPE_STRING,
                          GTK_ARG_READWRITE, ARG_NAME);

  gtkobject_class->set_arg = gst_artsdsink_set_arg;
  gtkobject_class->get_arg = gst_artsdsink_get_arg;

  gstelement_class->change_state = gst_artsdsink_change_state;
}

static void
gst_artsdsink_init(GstArtsdsink *artsdsink)
{
  artsdsink->sinkpad = gst_pad_new_from_template (sink_template, "sink");
  gst_element_add_pad(GST_ELEMENT(artsdsink), artsdsink->sinkpad);
  gst_pad_set_chain_function(artsdsink->sinkpad, gst_artsdsink_chain);

  artsdsink->connected = FALSE;
  artsdsink->mute = FALSE;

  // FIXME: get default from somewhere better than just putting them inline.
  artsdsink->signd = TRUE;
  artsdsink->depth = 16;
  artsdsink->channels = 2;
  artsdsink->frequency = 44100;
  artsdsink->connect_name = NULL;
}

static gboolean
gst_artsdsink_sync_parms (GstArtsdsink *artsdsink)
{
  g_return_val_if_fail (artsdsink != NULL, FALSE);
  g_return_val_if_fail (GST_IS_ARTSDSINK (artsdsink), FALSE);

  if (!artsdsink->connected) return TRUE;

  // Need to set stream to use new parameters: only way to do this is to reopen.
  gst_artsdsink_close_audio (artsdsink);
  return gst_artsdsink_open_audio (artsdsink);
}


static void
gst_artsdsink_chain (GstPad *pad, GstBuffer *buf)
{
  GstArtsdsink *artsdsink;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);

  artsdsink = GST_ARTSDSINK (gst_pad_get_parent (pad));

  if (GST_BUFFER_DATA (buf) != NULL) {
    gst_trace_add_entry(NULL, 0, buf, "artsdsink: writing to server");
    if (!artsdsink->mute && artsdsink->connected) {
      int bytes;
      void * bufptr = GST_BUFFER_DATA (buf);
      int bufsize = GST_BUFFER_SIZE (buf);
      GST_DEBUG (0, "artsdsink: stream=%p data=%p size=%d\n",
		 artsdsink->stream, GST_BUFFER_DATA (buf), GST_BUFFER_SIZE (buf));

      do {
	bytes = arts_write (artsdsink->stream, bufptr, bufsize);
	if(bytes < 0) {
	  fprintf(stderr,"arts_write error: %s\n", arts_error_text(bytes));
	  gst_buffer_unref (buf);
	  return;
	}
	bufptr += bytes;
	bufsize -= bytes;
      } while (bufsize > 0);
    }
  }
  gst_buffer_unref (buf);
}

static void
gst_artsdsink_set_arg (GtkObject *object,GtkArg *arg,guint id)
{
  GstArtsdsink *artsdsink;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_ARTSDSINK(object));
  artsdsink = GST_ARTSDSINK(object);

  switch(id) {
    case ARG_MUTE:
      artsdsink->mute = GTK_VALUE_BOOL(*arg);
      break;
    case ARG_DEPTH:
      artsdsink->depth = GTK_VALUE_ENUM (*arg);
      gst_artsdsink_sync_parms (artsdsink);
      break;
    case ARG_CHANNELS:
      artsdsink->channels = GTK_VALUE_ENUM (*arg);
      gst_artsdsink_sync_parms (artsdsink);
      break;
    case ARG_RATE:
      artsdsink->frequency = GTK_VALUE_INT (*arg);
      gst_artsdsink_sync_parms (artsdsink);
      break;
    case ARG_NAME:
      if (artsdsink->connect_name != NULL) g_free(artsdsink->connect_name);
      if (GTK_VALUE_STRING (*arg) == NULL)
	  artsdsink->connect_name = NULL;
      else
	  artsdsink->connect_name = g_strdup (GTK_VALUE_STRING (*arg));
      break;
    default:
      break;
  }
}

static void
gst_artsdsink_get_arg (GtkObject *object,GtkArg *arg,guint id)
{
  GstArtsdsink *artsdsink;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_ARTSDSINK(object));
  artsdsink = GST_ARTSDSINK(object);

  switch (id) {
    case ARG_MUTE:
      GTK_VALUE_BOOL(*arg) = artsdsink->mute;
      break;
    case ARG_DEPTH:
      GTK_VALUE_ENUM (*arg) = artsdsink->depth;
      break;
    case ARG_CHANNELS:
      GTK_VALUE_ENUM (*arg) = artsdsink->channels;
      break;
    case ARG_RATE:
      GTK_VALUE_INT (*arg) = artsdsink->frequency;
      break;
    case ARG_NAME:
      GTK_VALUE_STRING (*arg) = artsdsink->connect_name;
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}

static gboolean
plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *factory;

  factory = gst_elementfactory_new("artsdsink", GST_TYPE_ARTSDSINK,
				   &artsdsink_details);
  g_return_val_if_fail(factory != NULL, FALSE);

  sink_template = sink_factory ();
  gst_elementfactory_add_padtemplate(factory, sink_template);

  gst_plugin_add_factory(plugin, factory);

  return TRUE;
}

GstPluginDesc plugin_desc = {
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "artsdsink",
  plugin_init
};

static gboolean
gst_artsdsink_open_audio (GstArtsdsink *sink)
{
  const char * connname = "gstreamer";
  int errcode;

  // Name used by aRtsd for this connection.
  if (sink->connect_name != NULL) connname = sink->connect_name;

  // FIXME: this should only ever happen once per process.
  // Really, artsc needs to be made thread safe to fix this (and other related
  // problems).
  errcode = arts_init();
  if(errcode < 0) {
      fprintf(stderr,"arts_init error: %s\n", arts_error_text(errcode));
      return FALSE;
  }

  GST_DEBUG (0, "artsdsink: attempting to open connection to aRtsd server\n");
  sink->stream = arts_play_stream(sink->frequency, sink->depth,
				  sink->channels, connname);
  // FIXME: check connection
  //   GST_DEBUG (0, "artsdsink: can't open connection to aRtsd server\n");

  GST_FLAG_SET (sink, GST_ARTSDSINK_OPEN);
  sink->connected = TRUE;

  return TRUE;
}

static void
gst_artsdsink_close_audio (GstArtsdsink *sink)
{
  if (!sink->connected) return;

  arts_close_stream(sink->stream);
  arts_free();
  GST_FLAG_UNSET (sink, GST_ARTSDSINK_OPEN);
  sink->connected = FALSE;

  g_print("artsdsink: closed connection\n");
}

static GstElementStateReturn
gst_artsdsink_change_state (GstElement *element)
{
  g_return_val_if_fail (GST_IS_ARTSDSINK (element), FALSE);

  /* if going down into NULL state, close the stream if it's open */
  if (GST_STATE_PENDING (element) == GST_STATE_NULL) {
    if (GST_FLAG_IS_SET (element, GST_ARTSDSINK_OPEN))
      gst_artsdsink_close_audio (GST_ARTSDSINK (element));
    /* otherwise (READY or higher) we need to open the stream */
  } else {
    if (!GST_FLAG_IS_SET (element, GST_ARTSDSINK_OPEN)) {
      if (!gst_artsdsink_open_audio (GST_ARTSDSINK (element)))
	return GST_STATE_FAILURE;
    }
  }

  if (GST_ELEMENT_CLASS (parent_class)->change_state)
    return GST_ELEMENT_CLASS (parent_class)->change_state (element);
  return GST_STATE_SUCCESS;
}

