/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <string.h>

//#define DEBUG_ENABLED
#include <gstv4lsrc.h>

#include <linux/videodev.h>

static GstElementDetails gst_v4lsrc_details = {
  "Video (v4l) Source",
  "Source/Video",
  "Read from a Video for Linux capture device",
  VERSION,
  "Wim Taymans <wim.taymans@tvd.be>",
  "(C) 2000",
};

/* V4lSrc signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_WIDTH,
  ARG_HEIGHT,
  ARG_FORMAT,
  ARG_TUNE,
  ARG_TUNED,
  ARG_INPUT,
  ARG_NORM,
  ARG_VOLUME,
  ARG_MUTE,
  ARG_AUDIO_MODE,
  ARG_COLOR,
  ARG_BRIGHT,
  ARG_HUE,
  ARG_CONTRAST,
  ARG_DEVICE,
};


static void			gst_v4lsrc_class_init	(GstV4lSrcClass *klass);
static void			gst_v4lsrc_init		(GstV4lSrc *v4lsrc);

static void			gst_v4lsrc_set_arg	(GtkObject *object,GtkArg *arg,guint id);
static void			gst_v4lsrc_get_arg	(GtkObject *object,GtkArg *arg,guint id);

static GstElementStateReturn	gst_v4lsrc_change_state	(GstElement *element);
static void			gst_v4lsrc_close_v4l	(GstV4lSrc *src);
static gboolean			gst_v4lsrc_open_v4l	(GstV4lSrc *src);

static GstBuffer*		gst_v4lsrc_get		(GstPad *pad);
static GstPadNegotiateReturn 	gst_v4lsrc_negotiate 	(GstPad *pad, GstCaps **caps, gpointer *user_data);

static gboolean			gst_v4lsrc_sync_parms	(GstV4lSrc *v4lsrc);

static GstElementClass *parent_class = NULL;
////static guint gst_v4lsrc_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_v4lsrc_get_type (void)
{
  static GtkType v4lsrc_type = 0;

  if (!v4lsrc_type) {
    static const GtkTypeInfo v4lsrc_info = {
      "GstV4lSrc",
      sizeof(GstV4lSrc),
      sizeof(GstV4lSrcClass),
      (GtkClassInitFunc)gst_v4lsrc_class_init,
      (GtkObjectInitFunc)gst_v4lsrc_init,
      (GtkArgSetFunc)gst_v4lsrc_set_arg,
      (GtkArgGetFunc)gst_v4lsrc_get_arg,
      (GtkClassInitFunc)NULL,
    };
    v4lsrc_type = gtk_type_unique(GST_TYPE_ELEMENT,&v4lsrc_info);
  }
  return v4lsrc_type;
}

static void
gst_v4lsrc_class_init (GstV4lSrcClass *klass)
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtk_object_add_arg_type("GstV4lSrc::width", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_WIDTH);
  gtk_object_add_arg_type("GstV4lSrc::height", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_HEIGHT);
  gtk_object_add_arg_type("GstV4lSrc::format", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_FORMAT);

  gtk_object_add_arg_type("GstV4lSrc::tune", GTK_TYPE_ULONG,
                          GTK_ARG_WRITABLE, ARG_TUNE);
  gtk_object_add_arg_type("GstV4lSrc::tuned", GTK_TYPE_BOOL,
                          GTK_ARG_READABLE, ARG_TUNED);
  gtk_object_add_arg_type("GstV4lSrc::input", GTK_TYPE_INT,
                          GTK_ARG_WRITABLE, ARG_INPUT);
  gtk_object_add_arg_type("GstV4lSrc::norm", GTK_TYPE_INT,
                          GTK_ARG_WRITABLE, ARG_NORM);
  gtk_object_add_arg_type("GstV4lSrc::volume", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_VOLUME);
  gtk_object_add_arg_type("GstV4lSrc::mute", GTK_TYPE_BOOL,
                          GTK_ARG_READWRITE, ARG_MUTE);
  gtk_object_add_arg_type("GstV4lSrc::mode", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_AUDIO_MODE);
  gtk_object_add_arg_type("GstV4lSrc::color", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_COLOR);
  gtk_object_add_arg_type("GstV4lSrc::bright", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_BRIGHT);
  gtk_object_add_arg_type("GstV4lSrc::hue", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_HUE);
  gtk_object_add_arg_type("GstV4lSrc::contrast", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_CONTRAST);
  gtk_object_add_arg_type("GstV4lSrc::device", GTK_TYPE_STRING,
                          GTK_ARG_READWRITE, ARG_DEVICE);

  gtkobject_class->set_arg = gst_v4lsrc_set_arg;
  gtkobject_class->get_arg = gst_v4lsrc_get_arg;

  gstelement_class->change_state = gst_v4lsrc_change_state;
}

static void
gst_v4lsrc_init (GstV4lSrc *v4lsrc)
{
  v4lsrc->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(v4lsrc),v4lsrc->srcpad);

  gst_pad_set_get_function (v4lsrc->srcpad,gst_v4lsrc_get);
  gst_pad_set_negotiate_function (v4lsrc->srcpad,gst_v4lsrc_negotiate);

  /* if the destination cannot say what it wants, we give this */
  v4lsrc->width = 100;
  v4lsrc->height = 100;
  v4lsrc->format = -1;
  v4lsrc->buffer_size = v4lsrc->width * v4lsrc->height * 3;
  // make a grbber
  v4lsrc->grabber = grab_init();
  v4lsrc->device = NULL;
}

static GstPadNegotiateReturn
gst_v4lsrc_negotiate (GstPad *pad, GstCaps **caps, gpointer *user_data) 
{
  GstV4lSrc *v4lsrc;

  GST_DEBUG (0, "v4lsrc: negotiate %p\n", user_data); 

  v4lsrc = GST_V4LSRC (gst_pad_get_parent (pad));

  if (!*caps) {
    return GST_PAD_NEGOTIATE_FAIL;
  }
  else {
    gint width, height;
    gulong format;

    GST_DEBUG (0, "%08lx\n", gst_caps_get_fourcc_int (*caps, "format"));

    width  = gst_caps_get_int (*caps, "width");
    height = gst_caps_get_int (*caps, "height");

    format =  gst_caps_get_fourcc_int (*caps, "format");

    g_print ("v4lsrc: got format %08lx\n", format);

    switch (format) {
      case GST_MAKE_FOURCC ('R','G','B',' '):
      {
        gint depth, endianness, bpp;

        depth = gst_caps_get_int (*caps, "depth");
        bpp = gst_caps_get_int (*caps, "bpp");
        endianness = gst_caps_get_int (*caps, "endianness");

        GST_DEBUG (0, "%d\n", depth);
        g_print ("v4lsrc: got depth %d, bpp %d, endianness %d\n", depth, bpp, endianness);
	switch (depth) {
          case 15:
            v4lsrc->format = (endianness == G_LITTLE_ENDIAN ? 
      		     VIDEO_RGB15_LE:
      		     VIDEO_RGB15_BE);
            v4lsrc->buffer_size = width * height * 2;
            break;
          case 16:
            v4lsrc->format = (endianness == G_LITTLE_ENDIAN ? 
      		     VIDEO_RGB16_LE:
      		     VIDEO_RGB16_BE);
            v4lsrc->buffer_size = width * height * 2;
            break;
          case 24:
            v4lsrc->format = (endianness == G_LITTLE_ENDIAN ? 
      		     VIDEO_BGR24:
      		     VIDEO_RGB24);
            v4lsrc->buffer_size = width * height * 3;
            break;
          case 32:
            v4lsrc->format = (endianness == G_LITTLE_ENDIAN ? 
      		     VIDEO_BGR32:
      		     VIDEO_RGB32);
            v4lsrc->buffer_size = width * height * 4;
            break;
          default:
            *caps = NULL;
            return GST_PAD_NEGOTIATE_TRY;
	}
	break;
      }
      case GST_MAKE_FOURCC ('I','4','2','0'):
        v4lsrc->format = VIDEO_YUV420P;
        v4lsrc->buffer_size = width * height +
        		      width * height / 2;
	break;
      case GST_MAKE_FOURCC ('U','Y','V','Y'):
	if (G_BYTE_ORDER == G_BIG_ENDIAN) {
          v4lsrc->format = VIDEO_YUV422;
          v4lsrc->buffer_size = width * height * 2;
	  break;
	}
	else {
          *caps = NULL;
          return GST_PAD_NEGOTIATE_TRY;
	}
      case GST_MAKE_FOURCC ('Y','U','Y','2'):
	if (G_BYTE_ORDER == G_LITTLE_ENDIAN) {
          v4lsrc->format = VIDEO_YUV422;
          v4lsrc->buffer_size = width * height * 2;
	  break;
	}
	else {
          *caps = NULL;
          return GST_PAD_NEGOTIATE_TRY;
	}
      default:
        *caps = NULL;
        return GST_PAD_NEGOTIATE_TRY;
    }
    v4lsrc->width  = width;
    v4lsrc->height = height;

    if (gst_v4lsrc_sync_parms (v4lsrc)) {
      return GST_PAD_NEGOTIATE_AGREE;
    }
    else {
      *caps = NULL;
      return GST_PAD_NEGOTIATE_TRY;
    }
  }

  return GST_PAD_NEGOTIATE_FAIL;
}

static GstBuffer*
gst_v4lsrc_get (GstPad *pad)
{
  GstV4lSrc *v4lsrc;
  GstBuffer *buf = NULL;
  guint8 *grab_buf;

  g_return_val_if_fail (pad != NULL, NULL);

  v4lsrc = GST_V4LSRC (gst_pad_get_parent (pad));

  if (!gst_pad_get_caps (v4lsrc->srcpad) && 
      !gst_pad_renegotiate (v4lsrc->srcpad)) {
    return NULL;
  }

  buf = gst_buffer_new();
  GST_BUFFER_DATA(buf) = g_malloc(v4lsrc->buffer_size);
  GST_BUFFER_SIZE(buf) = v4lsrc->buffer_size;
  GST_DEBUG (0,"v4lsrc: making new buffer %p\n", GST_BUFFER_DATA(buf));

  GST_DEBUG (0,"v4lsrc: request buffer\n");
  // request a buffer from the grabber
  grab_buf = v4lsrc->grabber->grab_capture(v4lsrc->grabber, 0);
  //meta_pull->overlay_info->did_overlay = FALSE;

  g_assert(buf != NULL);

  GST_DEBUG (0,"v4lsrc: sending %d bytes in %p\n", GST_BUFFER_SIZE(buf), GST_BUFFER_DATA(buf));
  // copy the buffer
  memcpy(GST_BUFFER_DATA(buf), grab_buf, GST_BUFFER_SIZE(buf));

  GST_DEBUG (0,"v4lsrc: sent %d bytes in %p\n", GST_BUFFER_SIZE(buf), GST_BUFFER_DATA(buf));

  return buf;
}

static void
gst_v4lsrc_set_arg (GtkObject *object, GtkArg *arg, guint id)
{
  GstV4lSrc *src;
  int ret = 0;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_V4LSRC(object));
  src = GST_V4LSRC(object);

  switch (id) {
    case ARG_WIDTH:
      src->width = GTK_VALUE_INT(*arg);
      gst_v4lsrc_sync_parms(src);
      break;
    case ARG_HEIGHT:
      src->height = GTK_VALUE_INT(*arg);
      gst_v4lsrc_sync_parms(src);
      break;
    case ARG_FORMAT:
      src->format = GTK_VALUE_INT(*arg);
      break;
    case ARG_TUNE:
      src->tune = GTK_VALUE_ULONG(*arg);
      ret = src->grabber->grab_tune(src->grabber, src->tune);
      break;
    case ARG_INPUT:
      src->input = GTK_VALUE_INT(*arg);
      ret = src->grabber->grab_input(src->grabber, src->input, -1);
      break;
    case ARG_NORM:
      src->norm = GTK_VALUE_INT(*arg);
      ret = src->grabber->grab_input(src->grabber, -1, src->norm);
      break;
    case ARG_VOLUME:
      src->volume = GTK_VALUE_INT(*arg);
      ret = src->grabber->grab_setattr(src->grabber, GRAB_ATTR_VOLUME, src->volume);
      break;
    case ARG_MUTE:
      src->mute = GTK_VALUE_BOOL(*arg);
      ret = src->grabber->grab_setattr(src->grabber, GRAB_ATTR_MUTE, src->mute);
      break;
    case ARG_AUDIO_MODE:
      src->audio_mode = GTK_VALUE_INT(*arg);
      ret = src->grabber->grab_setattr(src->grabber, GRAB_ATTR_MODE, src->audio_mode);
      break;
    case ARG_COLOR:
      src->color = GTK_VALUE_INT(*arg);
      ret = src->grabber->grab_setattr(src->grabber, GRAB_ATTR_COLOR, src->color);
      break;
    case ARG_BRIGHT:
      src->bright = GTK_VALUE_INT(*arg);
      ret = src->grabber->grab_setattr(src->grabber, GRAB_ATTR_BRIGHT, src->bright);
      break;
    case ARG_HUE:
      src->hue = GTK_VALUE_INT(*arg);
      ret = src->grabber->grab_setattr(src->grabber, GRAB_ATTR_HUE, src->hue);
      break;
    case ARG_CONTRAST:
      src->contrast = GTK_VALUE_INT(*arg);
      ret = src->grabber->grab_setattr(src->grabber, GRAB_ATTR_CONTRAST, src->contrast);
      break;
    case ARG_DEVICE:
      if (src->device)
	g_free (src->device);
      src->device = g_strdup (GTK_VALUE_STRING (*arg));
      break;
    default:
      ret = -1;
      break;
  }
  if (ret == -1) {
    fprintf(stderr, "v4lsrc: error setting property\n");
  }
}

static void
gst_v4lsrc_get_arg (GtkObject *object, GtkArg *arg, guint id)
{
  GstV4lSrc *src;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_V4LSRC(object));
  src = GST_V4LSRC(object);

  g_print ("get arg\n");

  switch (id) {
    case ARG_WIDTH:
      GTK_VALUE_INT(*arg) = src->width;
      break;
    case ARG_HEIGHT:
      GTK_VALUE_INT(*arg) = src->height;
      break;
    case ARG_TUNED:
      GTK_VALUE_BOOL(*arg) = src->grabber->grab_tuned(src->grabber);
      break;
    case ARG_VOLUME:
      GTK_VALUE_INT(*arg) = src->grabber->grab_getattr(src->grabber, GRAB_ATTR_VOLUME);
      break;
    case ARG_MUTE:
      GTK_VALUE_INT(*arg) = src->grabber->grab_getattr(src->grabber, GRAB_ATTR_MUTE);
      break;
    case ARG_AUDIO_MODE:
      GTK_VALUE_INT(*arg) = src->grabber->grab_getattr(src->grabber, GRAB_ATTR_MODE);
      break;
    case ARG_COLOR:
      GTK_VALUE_INT(*arg) = src->grabber->grab_getattr(src->grabber, GRAB_ATTR_COLOR);
      break;
    case ARG_BRIGHT:
      GTK_VALUE_INT(*arg) = src->grabber->grab_getattr(src->grabber, GRAB_ATTR_BRIGHT);
      break;
    case ARG_HUE:
      GTK_VALUE_INT(*arg) = src->grabber->grab_getattr(src->grabber, GRAB_ATTR_HUE);
      break;
    case ARG_CONTRAST:
      GTK_VALUE_INT(*arg) = src->grabber->grab_getattr(src->grabber, GRAB_ATTR_CONTRAST);
      break;
    case ARG_DEVICE:
      GTK_VALUE_STRING(*arg) = src->device;
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}

static GstElementStateReturn
gst_v4lsrc_change_state (GstElement *element)
{
  g_return_val_if_fail(GST_IS_V4LSRC(element), FALSE);

  /* if going down into NULL state, close the file if it's open */
  if (GST_STATE_PENDING(element) == GST_STATE_NULL) {
    if (GST_FLAG_IS_SET(element,GST_V4LSRC_OPEN))
      gst_v4lsrc_close_v4l(GST_V4LSRC(element));
  /* otherwise (READY or higher) we need to open the sound card */
  } else {
    gst_info ("v4lsrc: opening\n");
    if (!GST_FLAG_IS_SET(element,GST_V4LSRC_OPEN)) {
      if (!gst_v4lsrc_open_v4l(GST_V4LSRC(element))) {
	gst_info ("v4lsrc: open failed\n");
        return GST_STATE_FAILURE;
      }
    }
  }

  if (GST_ELEMENT_CLASS(parent_class)->change_state)
    return GST_ELEMENT_CLASS(parent_class)->change_state(element);

  return GST_STATE_SUCCESS;
}

static gboolean
gst_v4lsrc_sync_parms (GstV4lSrc *src)
{
  gint linelength;
  gboolean success;

  g_return_val_if_fail(src != NULL, FALSE);
  g_return_val_if_fail(GST_IS_V4LSRC(src), FALSE);

  GST_DEBUG (0,"v4lsrc: resync %d %d %d\n", src->width, src->height, src->format);

  if (!src->grabber->opened) 
    return FALSE;

  if (src->grabber->grab_setparams(src->grabber, src->format, &src->width, &src->height, &linelength) != 0) {
    fprintf(stderr, "v4lsrc: error setting params\n");
    success = FALSE;
  }
  else {
    GST_DEBUG (0,"v4lsrc: resynced to %d %d %d\n", src->width, src->height, src->buffer_size);
    success = TRUE;
  }
  return success;
}

static gboolean
gst_v4lsrc_open_v4l (GstV4lSrc *src)
{
  g_return_val_if_fail(src->grabber != NULL, FALSE);
  g_return_val_if_fail(!src->grabber->opened, FALSE);

  if (src->grabber->grab_open(src->grabber, src->device) != -1) {
    gst_v4lsrc_sync_parms(src);
    GST_FLAG_SET(src, GST_V4LSRC_OPEN);
    return TRUE;
  }
  return FALSE;
}

static void
gst_v4lsrc_close_v4l (GstV4lSrc *src)
{
  g_return_if_fail(src->grabber != NULL);
  g_return_if_fail(src->grabber->opened);

  src->grabber->grab_close(src->grabber);
  GST_FLAG_UNSET(src, GST_V4LSRC_OPEN);
}

static gboolean
plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *factory;

  /* create an elementfactory for the v4lsrcparse element */
  factory = gst_elementfactory_new("v4lsrc",GST_TYPE_V4LSRC,
                                   &gst_v4lsrc_details);
  g_return_val_if_fail(factory != NULL, FALSE);
  gst_plugin_add_factory(plugin,factory);

  return TRUE;
}

GstPluginDesc plugin_desc = {
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "v4lsrc",
  plugin_init
};

