/* Gnome-Streamer
 * Copyright (C) <2001> Steve Baker <stevebaker_org@yahoo.co.uk>
 *
 * mono2stereo.c
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <mono2stereo.h>


static GstElementDetails mono2stereo_details = {
  "Mono to Stereo effect",
  "Filter/Effect",
  "Take a single channel and send it over 2 channels with pan",
  VERSION,
  "Steve Baker <stevebaker_org@yahoo.co.uk>",
  "(C) 2001",
};


/* Stereo signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_PAN,
};

static GstPadTemplate* 
mono2stereo_src_factory (void)
{
  return 
    gst_padtemplate_new (
    "src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    gst_caps_new (
    "sinesrc_src",
      "audio/raw",
      gst_props_new (
        "format",   GST_PROPS_STRING ("int"),
        "channels", GST_PROPS_INT (2),
      NULL)),
    NULL);
}

static GstPadTemplate* 
mono2stereo_sink_factory (void)
{
  return 
    gst_padtemplate_new (
      "sink",
      GST_PAD_SINK,
      GST_PAD_ALWAYS,
      gst_caps_new (
      "sinesrc_sink",
        "audio/raw",
        gst_props_new (
          "format",   GST_PROPS_STRING ("int"),
          "channels", GST_PROPS_INT (1),
        NULL)),
      NULL);
}

static GstPadTemplate *srctempl, *sinktempl;

static void	gst_mono2stereo_class_init		(GstMono2StereoClass *klass);
static void	gst_mono2stereo_init			(GstMono2Stereo *plugin);
static inline void gst_mono2stereo_update_pan(GstMono2Stereo *plugin);
static void	gst_mono2stereo_set_arg		(GtkObject *object, GtkArg *arg, guint id);
static void	gst_mono2stereo_get_arg		(GtkObject *object, GtkArg *arg, guint id);

static void	gst_mono2stereo_chain		(GstPad *pad, GstBuffer *buf);

static GstElementClass *parent_class = NULL;

static GstPadNegotiateReturn
gst_mono2stereo_negotiate_src (GstPad *pad, GstCaps **caps, gpointer *data)
{
  GstMono2Stereo* plugin=GST_MONO2STEREO (GST_OBJECT_PARENT (pad));
  if (*caps==NULL) return GST_PAD_NEGOTIATE_FAIL;
  *caps = gst_caps_copy_on_write (*caps);
  gst_caps_set(*caps,"channels",GST_PROPS_INT(1));
  plugin->width=gst_caps_get_int (*caps, "width");
  return gst_pad_negotiate_proxy(pad,plugin->sinkpad,caps);
}

static GstPadNegotiateReturn
gst_mono2stereo_negotiate_sink (GstPad *pad, GstCaps **caps, gpointer *data)
{
  GstMono2Stereo* plugin=GST_MONO2STEREO (GST_OBJECT_PARENT (pad));
  if (*caps==NULL) return GST_PAD_NEGOTIATE_FAIL;
  *caps = gst_caps_copy_on_write (*caps);
  gst_caps_set(*caps,"channels",GST_PROPS_INT(2));
  plugin->width=gst_caps_get_int (*caps, "width");
  return gst_pad_negotiate_proxy(pad,plugin->srcpad,caps);
}	

GtkType
gst_mono2stereo_get_type(void) {
  static GtkType mono2stereo_type = 0;

  if (!mono2stereo_type) {
    static const GtkTypeInfo mono2stereo_info = {
      "GstMono2Stereo",
      sizeof(GstMono2Stereo),
      sizeof(GstMono2StereoClass),
      (GtkClassInitFunc)gst_mono2stereo_class_init,
      (GtkObjectInitFunc)gst_mono2stereo_init,
      (GtkArgSetFunc)gst_mono2stereo_set_arg,
      (GtkArgGetFunc)gst_mono2stereo_get_arg,
      (GtkClassInitFunc)NULL,
    };
    mono2stereo_type = gtk_type_unique(GST_TYPE_ELEMENT,&mono2stereo_info);
  }
  return mono2stereo_type;
}

static void
gst_mono2stereo_class_init (GstMono2StereoClass *klass)
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtk_object_add_arg_type("GstMono2Stereo::pan", GTK_TYPE_FLOAT,
                          GTK_ARG_READWRITE, ARG_PAN);

  gtkobject_class->set_arg = gst_mono2stereo_set_arg;
  gtkobject_class->get_arg = gst_mono2stereo_get_arg;
}

static void
gst_mono2stereo_init (GstMono2Stereo *plugin)
{
  plugin->sinkpad = gst_pad_new_from_template(sinktempl,"sink");
  plugin->srcpad = gst_pad_new_from_template(srctempl,"src");

  gst_element_add_pad(GST_ELEMENT(plugin),plugin->sinkpad);
  gst_pad_set_chain_function(plugin->sinkpad,gst_mono2stereo_chain);
  gst_element_add_pad(GST_ELEMENT(plugin),plugin->srcpad);

  gst_pad_set_negotiate_function(plugin->sinkpad, gst_mono2stereo_negotiate_sink);
  gst_pad_set_negotiate_function(plugin->srcpad, gst_mono2stereo_negotiate_src);
  
  plugin->pan=0.0;
  gst_mono2stereo_update_pan(plugin);
}

static inline void gst_mono2stereo_update_pan(GstMono2Stereo *plugin)
{
  plugin->pan_right = (plugin->pan+1.0) / 2.0;
  plugin->pan_left = 1.0 - plugin->pan_right;
  g_print("%f %f %f\n", plugin->pan, plugin->pan_left, plugin->pan_right);
}

static void
gst_mono2stereo_chain (GstPad *pad,GstBuffer *buf_in)
{
  GstMono2Stereo *plugin;
  GstBuffer *buf_out;
  gint16 *data_in, *data_out;
  gint i;
  gint num_frames;
  
  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf_in != NULL);

  plugin = GST_MONO2STEREO(GST_OBJECT_PARENT (pad));
  g_return_if_fail(plugin != NULL);
  g_return_if_fail(GST_IS_MONO2STEREO(plugin));

  data_in = (gint16 *)GST_BUFFER_DATA(buf_in);
  num_frames = GST_BUFFER_SIZE(buf_in)/2;

  buf_out = gst_buffer_new();
  g_return_if_fail (buf_out);
  data_out = g_new(gint16, num_frames * 2);
  GST_BUFFER_DATA(buf_out) = (gpointer) data_out;
  GST_BUFFER_SIZE(buf_out) = num_frames * 4;

  for (i = 0; i < num_frames; i++) {
    data_out[i*2] = data_in[i] * plugin->pan_left;
    data_out[i*2+1] = data_in[i] * plugin->pan_right;
  }
  
  gst_buffer_unref(buf_in);
  gst_pad_push(plugin->srcpad,buf_out);
}

static void
gst_mono2stereo_set_arg (GtkObject *object, GtkArg *arg, guint id)
{
  GstMono2Stereo *plugin;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_MONO2STEREO(object));
  plugin = GST_MONO2STEREO(object);

  switch(id) {
    case ARG_PAN:
      if (GTK_VALUE_DOUBLE(*arg) < -1.0){
        plugin->pan = -1.0;
      } else if (GTK_VALUE_DOUBLE(*arg) > 1.0){
        plugin->pan = 1.0;
      } else {
        plugin->pan = GTK_VALUE_FLOAT(*arg);
      }
      gst_mono2stereo_update_pan(plugin);
      break;
    default:
      break;
  }
}

static void
gst_mono2stereo_get_arg (GtkObject *object, GtkArg *arg, guint id)
{
  GstMono2Stereo *plugin;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_MONO2STEREO(object));
  plugin = GST_MONO2STEREO(object);

  switch (id) {
    case ARG_PAN:
      GTK_VALUE_FLOAT(*arg) = plugin->pan;
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}

static gboolean
plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *factory;

  factory = gst_elementfactory_new("mono2stereo",GST_TYPE_MONO2STEREO,
                                   &mono2stereo_details);
  g_return_val_if_fail(factory != NULL, FALSE);
  
  srctempl = mono2stereo_src_factory();
  gst_elementfactory_add_padtemplate (factory, srctempl);

  sinktempl = mono2stereo_sink_factory();
  gst_elementfactory_add_padtemplate (factory, sinktempl);
  
  gst_plugin_add_factory(plugin,factory);

  return TRUE;
}

GstPluginDesc plugin_desc = {
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "mono2stereo",
  plugin_init
};